@@ -12,7 +12,10 @@
 #include <array>
 #include <cerrno>
 #include <cstring>
+#include <filesystem>
 #include <memory>
+#include <ranges>
+#include <span>
 
 #include <QCoreApplication>
 #include <QDateTime>
@@ -28,26 +31,38 @@
 #include <KUser>
 
 #include <KConfigGroup>
+#include <KIO/AuthInfo>
 #include <KLocalizedString>
 #include <kio/ioworker_defaults.h>
 
 #ifdef Q_OS_WIN
-#include <filesystem> // for permissions
-using namespace std::filesystem;
 #include <qplatformdefs.h>
-#else
-#include <utime.h>
 #endif
 
 #include "kio_sftp_debug.h"
 #include "kio_sftp_trace_debug.h"
 
-// For MinGW compatibility
-#ifndef QT_STAT_LNK
-#define QT_STAT_LNK 0120000
-#endif // QT_STAT_LNK
+using namespace KIO;
+using namespace std::filesystem;
 
-#define KIO_SFTP_SPECIAL_TIMEOUT 30
+namespace std
+{
+template<>
+struct default_delete<struct sftp_file_struct> {
+    void operator()(struct sftp_file_struct *ptr) const
+    {
+        if (ptr) {
+            sftp_close(ptr);
+        }
+    }
+};
+} // namespace std
+
+using UniqueSFTPFilePtr = std::unique_ptr<struct sftp_file_struct>;
+
+namespace
+{
+constexpr auto KIO_SFTP_SPECIAL_TIMEOUT_MS = 30;
 
 // How big should each data packet be? Definitely not bigger than 64kb or
 // you will overflow the 2 byte size variable in a sftp packet.
@@ -67,42 +82,16 @@ using namespace std::filesystem;
 // [https://bugs.kde.org/show_bug.cgi?id=404890]) we ought to be more conservative!
 // At the same time there's no bug reports about the 60k requests being too large so
 // perhaps all popular servers effectively support at least 64k.
-#define MAX_XFER_BUF_SIZE (60 * 1024)
+constexpr auto MAX_XFER_BUF_SIZE = (60ULL * 1024);
 
-#define KSFTP_ISDIR(sb) (sb->type == SSH_FILEXFER_TYPE_DIRECTORY)
-
-// Pseudo plugin class to embed meta data
-class KIOPluginForMetaData : public QObject
+inline bool KSFTP_ISDIR(SFTPAttributesPtr &sb)
 {
-    Q_OBJECT
-    Q_PLUGIN_METADATA(IID "org.kde.kio.worker.sftp" FILE "sftp.json")
-};
-
-using namespace KIO;
-extern "C" {
-int Q_DECL_EXPORT kdemain(int argc, char **argv)
-{
-    QCoreApplication app(argc, argv);
-    app.setApplicationName("kio_sftp");
-
-    qCDebug(KIO_SFTP_LOG) << "*** Starting kio_sftp ";
-
-    if (argc != 4) {
-        qCDebug(KIO_SFTP_LOG) << "Usage: kio_sftp protocol domain-socket1 domain-socket2";
-        exit(-1);
-    }
-
-    SFTPWorker worker(argv[2], argv[3]);
-    worker.dispatchLoop();
-
-    qCDebug(KIO_SFTP_LOG) << "*** kio_sftp Done";
-    return 0;
-}
+    return sb->type == SSH_FILEXFER_TYPE_DIRECTORY;
 }
 
 // Converts SSH error into KIO error. Only must be called for error handling
 // as this will always return an error state and never NoError.
-static int toKIOError(const int err)
+int toKIOError(const int err)
 {
     switch (err) {
     case SSH_FX_NO_SUCH_FILE:
@@ -129,16 +118,15 @@ static int toKIOError(const int err)
     return KIO::ERR_UNKNOWN;
 }
 
-// Writes 'len' bytes from 'buf' to the file handle 'fd'.
-static int writeToFile(int fd, const char *buf, int len)
+// Writes buf into fd.
+int writeToFile(int fd, const std::span<const char> &buf)
 {
-    while (len > 0) {
-        const auto lenSize = static_cast<size_t>(len); // len is always >> 0 and, being an int, always << size_t
-        const ssize_t written = write(fd, buf, lenSize);
+    size_t offset = 0;
+    while (offset != buf.size()) {
+        const auto written = write(fd, &buf[offset], buf.size() - offset);
 
         if (written >= 0) {
-            buf += written;
-            len -= written;
+            offset += written;
             continue;
         }
 
@@ -157,7 +145,7 @@ static int writeToFile(int fd, const char *buf, int len)
     return 0;
 }
 
-static bool wasUsernameChanged(const QString &username, const KIO::AuthInfo &info)
+bool wasUsernameChanged(const QString &username, const KIO::AuthInfo &info)
 {
     QString loginName(username);
     // If username is empty, assume the current logged in username. Why ?
@@ -172,7 +160,7 @@ static bool wasUsernameChanged(const QString &username, const KIO::AuthInfo &inf
 }
 
 // The callback function for libssh
-static int auth_callback(const char *prompt, char *buf, size_t len, int echo, int verify, void *userdata)
+int auth_callback(const char *prompt, char *buf, size_t len, int echo, int verify, void *userdata)
 {
     if (userdata == nullptr) {
         return -1;
@@ -186,7 +174,7 @@ static int auth_callback(const char *prompt, char *buf, size_t len, int echo, in
     return 0;
 }
 
-static void log_callback(int priority, const char *function, const char *buffer, void *userdata)
+void log_callback(int priority, const char *function, const char *buffer, void *userdata)
 {
     if (userdata == nullptr) {
         return;
@@ -196,6 +184,49 @@ static void log_callback(int priority, const char *function, const char *buffer,
     worker->log_callback(priority, function, buffer, userdata);
 }
 
+inline std::optional<perms> posixToOptionalPerms(uint32_t mode) noexcept
+{
+    if (mode < 0) {
+        return {};
+    }
+    return static_cast<perms>(mode) & perms::mask;
+}
+
+inline mode_t permsToPosix(perms mode) noexcept
+{
+    return static_cast<mode_t>(mode);
+}
+} // namespace
+
+// Pseudo plugin class to embed meta data
+class KIOPluginForMetaData : public QObject
+{
+    Q_OBJECT
+    Q_PLUGIN_METADATA(IID "org.kde.kio.worker.sftp" FILE "sftp.json")
+};
+
+extern "C" {
+int Q_DECL_EXPORT kdemain(int argc, char **argv)
+{
+    QCoreApplication app(argc, argv);
+    app.setApplicationName("kio_sftp");
+
+    qCDebug(KIO_SFTP_LOG) << "*** Starting kio_sftp ";
+
+    if (argc != 4) {
+        qCWarning(KIO_SFTP_LOG) << "Usage: kio_sftp protocol domain-socket1 domain-socket2";
+        exit(-1);
+    }
+
+    std::span args{argv, std::make_unsigned_t<decltype(argc)>(argc)};
+    SFTPWorker worker(args[2], args[3]);
+    worker.dispatchLoop();
+
+    qCDebug(KIO_SFTP_LOG) << "*** kio_sftp Done";
+    return 0;
+}
+}
+
 SFTPWorker::SFTPWorker(const QByteArray &poolSocket, const QByteArray &appSocket)
     : WorkerBase(QByteArrayLiteral("kio_sftp"), poolSocket, appSocket)
 {
@@ -280,9 +311,9 @@ Result SFTPWorker::init()
     mCallbacks->userdata = this;
     mCallbacks->auth_function = ::auth_callback;
 
-    ssh_callbacks_init(mCallbacks)
+    ssh_callbacks_init(mCallbacks);
 
-        bool ok;
+    bool ok = false;
     int level = qEnvironmentVariableIntValue("KIO_SFTP_LOG_VERBOSITY", &ok);
     if (ok) {
         int rc = ssh_set_log_level(level);
@@ -317,7 +348,7 @@ int SFTPWorker::authenticateKeyboardInteractive(AuthInfo &info)
 
         for (int iInt = 0; iInt < n; ++iInt) {
             const auto i = static_cast<unsigned int>(iInt); // can only be >0
-            char echo;
+            char echo = 0;
             const char *answer = "";
 
             const QString prompt = QString::fromUtf8(ssh_userauth_kbdint_getprompt(mSession, i, &echo));
@@ -360,25 +391,25 @@ int SFTPWorker::authenticateKeyboardInteractive(AuthInfo &info)
                     return SSH_AUTH_ERROR;
                 }
                 break;
+            }
+
+            if (prompt.startsWith(QLatin1String("password:"), Qt::CaseInsensitive)) {
+                info.prompt = i18n("Please enter your password.");
             } else {
-                if (prompt.startsWith(QLatin1String("password:"), Qt::CaseInsensitive)) {
-                    info.prompt = i18n("Please enter your password.");
-                } else {
-                    info.prompt = prompt;
-                }
-                info.comment = info.url.url();
-                info.commentLabel = i18n("Site:");
-                info.setExtraField(QLatin1String("hide-username-line"), true);
+                info.prompt = prompt;
+            }
+            info.comment = info.url.url();
+            info.commentLabel = i18n("Site:");
+            info.setExtraField(QLatin1String("hide-username-line"), true);
 
-                if (openPasswordDialog(info) == KJob::NoError) {
-                    qCDebug(KIO_SFTP_LOG) << "Got the answer from the password dialog";
-                    answer = info.password.toUtf8().constData();
-                }
+            if (openPasswordDialog(info) == KJob::NoError) {
+                qCDebug(KIO_SFTP_LOG) << "Got the answer from the password dialog";
+                answer = info.password.toUtf8().constData();
+            }
 
-                if (ssh_userauth_kbdint_setanswer(mSession, i, answer) < 0) {
-                    qCDebug(KIO_SFTP_LOG) << "An error occurred setting the answer: " << ssh_get_error(mSession);
-                    return SSH_AUTH_ERROR;
-                }
+            if (ssh_userauth_kbdint_setanswer(mSession, i, answer) < 0) {
+                qCDebug(KIO_SFTP_LOG) << "An error occurred setting the answer: " << ssh_get_error(mSession);
+                return SSH_AUTH_ERROR;
             }
         }
         err = ssh_userauth_kbdint(mSession, nullptr, nullptr);
@@ -429,17 +460,13 @@ Result SFTPWorker::createUDSEntry(SFTPAttributesPtr sb, UDSEntry &entry, const Q
         }
     }
 
-    mode_t access = 0;
+    perms access = perms::none;
     long long fileType = QT_STAT_REG;
     uint64_t size = 0U;
     if (isBrokenLink) {
         // It is a link pointing to nowhere
         fileType = QT_STAT_MASK - 1;
-#ifdef Q_OS_WIN
-        access = static_cast<mode_t>(perms::owner_all | perms::group_all | perms::others_all);
-#else
-        access = S_IRWXU | S_IRWXG | S_IRWXO;
-#endif
+        access = perms::all;
         size = 0LL;
     } else {
         switch (sb->type) {
@@ -457,11 +484,11 @@ Result SFTPWorker::createUDSEntry(SFTPAttributesPtr sb, UDSEntry &entry, const Q
             fileType = QT_STAT_MASK - 1;
             break;
         }
-        access = sb->permissions & 07777;
+        access = posixToOptionalPerms(sb->permissions).value_or(perms::none);
         size = sb->size;
     }
     entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, fileType);
-    entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, access);
+    entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, permsToPosix(access));
     entry.fastInsert(KIO::UDSEntry::UDS_SIZE, size);
 
     if (details > 0) {
@@ -602,10 +629,8 @@ Result SFTPWorker::sftpOpenConnection(const AuthInfo &info)
 
     qCDebug(KIO_SFTP_LOG) << "Trying to connect to the SSH server";
 
-    unsigned int effectivePort;
-    if (mPort > 0) {
-        effectivePort = mPort;
-    } else {
+    unsigned int effectivePort = mPort;
+    if (effectivePort <= 0) {
         effectivePort = DEFAULT_SFTP_PORT;
         ssh_options_get_port(mSession, &effectivePort);
     }
@@ -932,7 +957,7 @@ Result SFTPWorker::openConnectionWithoutCloseOnError()
         mUsername = info.username;
     }
 
-    setTimeoutSpecialCommand(KIO_SFTP_SPECIAL_TIMEOUT);
+    setTimeoutSpecialCommand(KIO_SFTP_SPECIAL_TIMEOUT_MS);
 
     mConnected = true;
 
@@ -1001,7 +1026,7 @@ Result SFTPWorker::special(const QByteArray &)
                               << "- SSH errorString:" << ssh_get_error(mSession);
     }
 
-    setTimeoutSpecialCommand(KIO_SFTP_SPECIAL_TIMEOUT);
+    setTimeoutSpecialCommand(KIO_SFTP_SPECIAL_TIMEOUT_MS);
 
     return Result::pass();
 }
@@ -1055,7 +1080,7 @@ Result SFTPWorker::open(const QUrl &url, QIODevice::OpenMode mode)
     }
 
     if (flags & O_CREAT) {
-        mOpenFile = sftp_open(mSftp, path_c.constData(), flags, 0644);
+        mOpenFile = sftp_open(mSftp, path_c.constData(), flags, permsToPosix(perms::owner_read | perms::owner_write | perms::group_read | perms::others_read));
     } else {
         mOpenFile = sftp_open(mSftp, path_c.constData(), flags, 0);
     }
@@ -1113,10 +1138,15 @@ Result SFTPWorker::write(const QByteArray &data)
 
     Q_ASSERT(mOpenFile != nullptr);
 
-    if (!sftpWrite(mOpenFile, data, nullptr)) {
-        qCDebug(KIO_SFTP_LOG) << "Could not write to " << mOpenUrl;
-        (void)close();
-        return Result::fail(KIO::ERR_CANNOT_WRITE, mOpenUrl.toDisplayString());
+    for (const auto &response : asyncWrite(mOpenFile, [data]() -> QCoro::Generator<ReadResponse> {
+             co_yield {.filedata = data};
+         }())) {
+        if (response.error != KJob::NoError) {
+            qCDebug(KIO_SFTP_LOG) << "Could not write to " << mOpenUrl;
+            (void)close();
+            return Result::fail(KIO::ERR_CANNOT_WRITE, mOpenUrl.toDisplayString());
+        }
+        // We don't care about progress in this function. We'll emit a single written() once done.
     }
 
     written(data.size());
@@ -1150,7 +1180,7 @@ Result SFTPWorker::truncate(KIO::filesize_t length)
     SFTPAttributesPtr attr(sftp_fstat(mOpenFile));
     if (attr) {
         attr->size = length;
-        if (sftp_setstat(mSftp, mOpenUrl.path().toUtf8().constData(), attr.data()) == 0) {
+        if (sftp_setstat(mSftp, mOpenUrl.path().toUtf8().constData(), attr.get()) == 0) {
             truncated(length);
         } else {
             errorCode = toKIOError(sftp_get_error(mSftp));
@@ -1220,9 +1250,7 @@ Result SFTPWorker::sftpGet(const QUrl &url, KIO::fileoffset_t offset, int fd)
 
     QByteArray path = url.path().toUtf8();
 
-    sftp_file file = nullptr;
     KIO::filesize_t totalbytesread = 0;
-    QByteArray filedata;
 
     SFTPAttributesPtr sb(sftp_lstat(mSftp, path.constData()));
     if (sb == nullptr) {
@@ -1241,15 +1269,12 @@ Result SFTPWorker::sftpGet(const QUrl &url, KIO::fileoffset_t offset, int fd)
     }
 
     // Open file
-    file = sftp_open(mSftp, path.constData(), O_RDONLY, 0);
-    if (file == nullptr) {
+    const UniqueSFTPFilePtr file(sftp_open(mSftp, path.constData(), O_RDONLY, 0));
+    if (!file) {
         return Result::fail(KIO::ERR_CANNOT_OPEN_FOR_READING, url.toString());
     }
-    const auto fileFree = qScopeGuard([file] {
-        sftp_close(file);
-    });
 
-    if (const Result result = sftpSendMimetype(file, url); !result.success()) {
+    if (const Result result = sftpSendMimetype(file.get(), url); !result.success()) {
         return result;
     }
 
@@ -1270,42 +1295,27 @@ Result SFTPWorker::sftpGet(const QUrl &url, KIO::fileoffset_t offset, int fd)
 
     // If we can resume, offset the buffer properly.
     if (offset > 0 && ((unsigned long long)offset < sb->size)) {
-        if (sftp_seek64(file, offset) == 0) {
+        if (sftp_seek64(file.get(), offset) == 0) {
             canResume();
             totalbytesread = offset;
             qCDebug(KIO_SFTP_LOG) << "Resume offset: " << QString::number(offset);
         }
     }
 
-    SFTPWorker::GetRequest request(file, sb->size);
-
-    for (;;) {
-        // Enqueue get requests
-        if (!request.enqueueChunks()) {
-            return Result::fail(KIO::ERR_CANNOT_READ, url.toString());
+    auto reader = asyncRead(file.get(), sb->size);
+    for (const auto &response : reader) {
+        if (response.error != KJob::NoError) {
+            return Result::fail(response.error, url.toString());
         }
 
-        filedata.clear();
-        const auto bytesread = request.readChunks(filedata);
-        // Read pending get requests
-        if (bytesread == -1) {
-            return Result::fail(KIO::ERR_CANNOT_READ, url.toString());
-        } else if (bytesread == 0) {
-            if (file->eof)
-                break;
-            else
-                continue;
-        }
-
-        int error = KJob::NoError;
         if (fd == -1) {
-            data(filedata);
-        } else if ((error = writeToFile(fd, filedata.constData(), filedata.size())) != KJob::NoError) {
+            data(response.filedata);
+        } else if (int error = writeToFile(fd, response.filedata); error != KJob::NoError) {
             return Result::fail(error, url.toString());
         }
-        // increment total bytes read
-        totalbytesread += filedata.length();
 
+        // increment total bytes read
+        totalbytesread += response.filedata.length();
         processedSize(totalbytesread);
     }
 
@@ -1326,9 +1336,12 @@ Result SFTPWorker::put(const QUrl &url, int permissions, KIO::JobFlags flags)
     return sftpPut(url, permissions, flags);
 }
 
-Result SFTPWorker::sftpPut(const QUrl &url, int permissions, JobFlags flags, int fd)
+Result SFTPWorker::sftpPut(const QUrl &url, int permissionsMode, JobFlags flags, int fd)
 {
-    qCDebug(KIO_SFTP_LOG) << url << ", permissions =" << permissions << ", overwrite =" << (flags & KIO::Overwrite) << ", resume =" << (flags & KIO::Resume);
+    qCDebug(KIO_SFTP_LOG) << url << ", permissions =" << permissionsMode << ", overwrite =" << (flags & KIO::Overwrite)
+                          << ", resume =" << (flags & KIO::Resume);
+
+    auto permissions(posixToOptionalPerms(permissionsMode));
 
     const auto loginResult = sftpLogin();
     if (!loginResult.success()) {
@@ -1349,7 +1362,7 @@ Result SFTPWorker::sftpPut(const QUrl &url, int permissions, JobFlags flags, int
 
     // Don't change permissions of the original file
     if (bOrigExists) {
-        permissions = sb->permissions;
+        permissions = posixToOptionalPerms(sb->permissions);
         owner = sb->uid;
         group = sb->gid;
     }
@@ -1385,84 +1398,86 @@ Result SFTPWorker::sftpPut(const QUrl &url, int permissions, JobFlags flags, int
     }
 
     QByteArray dest;
-    int result = -1;
-    sftp_file file = nullptr;
+    if (bMarkPartial) {
+        qCDebug(KIO_SFTP_LOG) << "Appending .part extension to" << dest_orig;
+        dest = dest_part_c;
+        if (bPartExists && !(flags & KIO::Resume)) {
+            qCDebug(KIO_SFTP_LOG) << "Deleting partial file" << dest_part;
+            sftp_unlink(mSftp, dest_part_c.constData());
+            // Catch errors when we try to open the file.
+        }
+    } else {
+        dest = dest_orig_c; // Will be automatically truncated below...
+    } // bMarkPartial
+
+    UniqueSFTPFilePtr destFile{};
     KIO::fileoffset_t totalBytesSent = 0;
-
-    int errorCode = KJob::NoError;
-    // Loop until we got 0 (end of data)
-    do {
-        QByteArray buffer;
-
-        if (fd == -1) {
-            dataReq(); // Request for data
-            result = readData(buffer);
-            if (result < 0) {
-                qCDebug(KIO_SFTP_LOG) << "unexpected error during readData";
+    if ((flags & KIO::Resume)) {
+        qCDebug(KIO_SFTP_LOG) << "Trying to append: " << dest;
+        destFile.reset(sftp_open(mSftp, dest.constData(), O_RDWR, 0)); // append if resuming
+        if (destFile) {
+            SFTPAttributesPtr fstat(sftp_fstat(destFile.get()));
+            if (fstat) {
+                sftp_seek64(destFile.get(), fstat->size); // Seek to end TODO
+                totalBytesSent += fstat->size;
             }
-        } else {
-            char buf[MAX_XFER_BUF_SIZE]; //
-            result = ::read(fd, buf, sizeof(buf));
-            if (result < 0) {
-                qCDebug(KIO_SFTP_LOG) << "failed to read" << errno;
-                errorCode = ERR_CANNOT_READ;
-                break;
-            }
-            buffer = QByteArray(buf, result);
+        }
+    } else {
+        perms initialMode = perms::owner_read | perms::owner_write | perms::group_read | perms::others_read;
+        if (permissions.has_value()) {
+            initialMode = permissions.value() | perms::owner_write | perms::owner_read;
         }
 
-        if (result >= 0) {
-            if (dest.isEmpty()) {
-                if (bMarkPartial) {
-                    qCDebug(KIO_SFTP_LOG) << "Appending .part extension to" << dest_orig;
-                    dest = dest_part_c;
-                    if (bPartExists && !(flags & KIO::Resume)) {
-                        qCDebug(KIO_SFTP_LOG) << "Deleting partial file" << dest_part;
-                        sftp_unlink(mSftp, dest_part_c.constData());
-                        // Catch errors when we try to open the file.
-                    }
+        qCDebug(KIO_SFTP_LOG) << "Trying to open:" << QString(dest) << ", mode=" << QString::number(permsToPosix(initialMode));
+        destFile.reset(sftp_open(mSftp, dest.constData(), O_CREAT | O_TRUNC | O_WRONLY, permsToPosix(initialMode)));
+    } // flags & KIO::Resume
+
+    auto closeOnError = [&destFile, bMarkPartial, this, dest, url](int errorCode) -> Result {
+        qCDebug(KIO_SFTP_LOG) << "Error during 'put'. Aborting.";
+
+        if (destFile != nullptr) {
+            destFile = nullptr; // close to force a data flush before we stat
+
+            SFTPAttributesPtr attr(sftp_stat(mSftp, dest.constData()));
+            if (bMarkPartial && attr != nullptr) {
+                size_t size = configValue(QStringLiteral("MinimumKeepSize"), DEFAULT_MINIMUM_KEEP_SIZE);
+                if (attr->size < size) {
+                    sftp_unlink(mSftp, dest.constData());
+                }
+            }
+        }
+
+        return errorCode == KJob::NoError ? Result::pass() : Result::fail(errorCode, url.toString());
+    };
+
+    if (destFile == nullptr) {
+        qCDebug(KIO_SFTP_LOG) << "COULD NOT WRITE " << QString(dest) << ", permissions=" << permsToPosix(permissions.value_or(perms::none))
+                              << ", error=" << ssh_get_error(mSession);
+        if (sftp_get_error(mSftp) == SSH_FX_PERMISSION_DENIED) {
+            return closeOnError(KIO::ERR_WRITE_ACCESS_DENIED);
+        }
+        return closeOnError(KIO::ERR_CANNOT_OPEN_FOR_WRITING);
+    } // file
+
+    auto reader = [fd, this]() -> QCoro::Generator<ReadResponse> {
+        for (int result = 1; result > 0;) {
+            ReadResponse response;
+            if (fd == -1) {
+                dataReq(); // Request for data
+                result = readData(response.filedata);
+                if (result < 0) {
+                    qCDebug(KIO_SFTP_LOG) << "unexpected error during readData";
+                }
+            } else {
+                std::array<char, MAX_XFER_BUF_SIZE> buf{};
+                result = ::read(fd, buf.data(), buf.size());
+                if (result < 0) {
+                    qCDebug(KIO_SFTP_LOG) << "failed to read" << errno;
+                    response.error = ERR_CANNOT_READ;
                 } else {
-                    dest = dest_orig_c; // Will be automatically truncated below...
-                } // bMarkPartial
-
-                if ((flags & KIO::Resume)) {
-                    qCDebug(KIO_SFTP_LOG) << "Trying to append: " << dest;
-                    file = sftp_open(mSftp, dest.constData(), O_RDWR, 0); // append if resuming
-                    if (file) {
-                        SFTPAttributesPtr fstat(sftp_fstat(file));
-                        if (fstat) {
-                            sftp_seek64(file, fstat->size); // Seek to end TODO
-                            totalBytesSent += fstat->size;
-                        }
-                    }
-                } else {
-                    mode_t initialMode;
-
-                    if (permissions != -1) {
-#ifdef Q_OS_WIN
-                        initialMode = permissions | static_cast<mode_t>(perms::owner_write | perms::owner_read);
-#else
-                        initialMode = permissions | S_IWUSR | S_IRUSR;
-#endif
-                    } else {
-                        initialMode = 0644;
-                    }
-
-                    qCDebug(KIO_SFTP_LOG) << "Trying to open:" << QString(dest) << ", mode=" << QString::number(initialMode);
-                    file = sftp_open(mSftp, dest.constData(), O_CREAT | O_TRUNC | O_WRONLY, initialMode);
-                } // flags & KIO::Resume
-
-                if (file == nullptr) {
-                    qCDebug(KIO_SFTP_LOG) << "COULD NOT WRITE " << QString(dest) << ", permissions=" << permissions << ", error=" << ssh_get_error(mSession);
-                    if (sftp_get_error(mSftp) == SSH_FX_PERMISSION_DENIED) {
-                        errorCode = KIO::ERR_WRITE_ACCESS_DENIED;
-                    } else {
-                        errorCode = KIO::ERR_CANNOT_OPEN_FOR_WRITING;
-                    }
-                    result = -1;
-                    continue;
-                } // file
-            } // dest.isEmpty
+                    response.filedata = QByteArray(buf.data(), result);
+                }
+            }
 
             if (result == 0) {
                 // proftpd stumbles over zero size writes.
@@ -1475,43 +1490,25 @@ Result SFTPWorker::sftpPut(const QUrl &url, int permissions, JobFlags flags, int
                 break;
             }
 
-            bool success = sftpWrite(file, buffer, [this, &totalBytesSent](int bytes) {
-                totalBytesSent += bytes;
-                processedSize(totalBytesSent);
-            });
-            if (!success) {
-                qCDebug(KIO_SFTP_LOG) << "totalBytesSent at error:" << totalBytesSent;
-                errorCode = KIO::ERR_CANNOT_WRITE;
-                result = -1;
-                break;
-            }
-        } // result
-    } while (result > 0);
+            co_yield response;
+        }
+    };
 
-    // An error occurred deal with it.
-    if (result < 0) {
-        qCDebug(KIO_SFTP_LOG) << "Error during 'put'. Aborting.";
-
-        if (file != nullptr) {
-            sftp_close(file);
-
-            SFTPAttributesPtr attr(sftp_stat(mSftp, dest.constData()));
-            if (bMarkPartial && attr != nullptr) {
-                size_t size = configValue(QStringLiteral("MinimumKeepSize"), DEFAULT_MINIMUM_KEEP_SIZE);
-                if (attr->size < size) {
-                    sftp_unlink(mSftp, dest.constData());
-                }
-            }
+    for (const auto &response : asyncWrite(destFile.get(), reader())) {
+        if (response.error != KJob::NoError) {
+            qCDebug(KIO_SFTP_LOG) << "totalBytesSent at error:" << totalBytesSent;
+            return closeOnError(KIO::ERR_CANNOT_WRITE);
         }
 
-        return errorCode == KJob::NoError ? Result::pass() : Result::fail(errorCode, url.toString());
+        totalBytesSent += response.bytes;
+        processedSize(totalBytesSent);
     }
 
-    if (file == nullptr) { // we got nothing to write out, so we never opened the file
+    if (destFile == nullptr) { // we got nothing to write out, so we never opened the file
         return Result::pass();
     }
 
-    if (sftp_close(file) < 0) {
+    if (sftp_close(destFile.release()) < 0) {
         qCWarning(KIO_SFTP_LOG) << "Error when closing file descriptor";
         return Result::fail(KIO::ERR_CANNOT_WRITE, url.toString());
     }
@@ -1532,9 +1529,9 @@ Result SFTPWorker::sftpPut(const QUrl &url, int permissions, JobFlags flags, int
     }
 
     // set final permissions
-    if (permissions != -1 && !(flags & KIO::Resume)) {
-        qCDebug(KIO_SFTP_LOG) << "Trying to set final permissions of " << dest_orig << " to " << QString::number(permissions);
-        if (sftp_chmod(mSftp, dest_orig_c.constData(), permissions) < 0) {
+    if (permissions.has_value() && !(flags & KIO::Resume)) {
+        qCDebug(KIO_SFTP_LOG) << "Trying to set final permissions of " << dest_orig << " to " << QString::number(permsToPosix(permissions.value()));
+        if (sftp_chmod(mSftp, dest_orig_c.constData(), permsToPosix(permissions.value())) < 0) {
             warning(i18n("Could not change permissions for\n%1", url.toString()));
             return Result::pass();
         }
@@ -1554,7 +1551,7 @@ Result SFTPWorker::sftpPut(const QUrl &url, int permissions, JobFlags flags, int
     if (!mtimeStr.isEmpty()) {
         QDateTime dt = QDateTime::fromString(mtimeStr, Qt::ISODate);
         if (dt.isValid()) {
-            struct timeval times[2];
+            std::array<struct timeval, 2> times{};
 
             SFTPAttributesPtr attr(sftp_lstat(mSftp, dest_orig_c.constData()));
             if (attr != nullptr) {
@@ -1563,7 +1560,7 @@ Result SFTPWorker::sftpPut(const QUrl &url, int permissions, JobFlags flags, int
                 times[0].tv_usec = times[1].tv_usec = 0;
 
                 qCDebug(KIO_SFTP_LOG) << "Trying to restore mtime for " << dest_orig << " to: " << mtimeStr;
-                result = sftp_utimes(mSftp, dest_orig_c.constData(), times);
+                int result = sftp_utimes(mSftp, dest_orig_c.constData(), times.data());
                 if (result < 0) {
                     qCWarning(KIO_SFTP_LOG) << "Failed to set mtime for" << dest_orig;
                 }
@@ -1574,34 +1571,6 @@ Result SFTPWorker::sftpPut(const QUrl &url, int permissions, JobFlags flags, int
     return Result::pass();
 }
 
-bool SFTPWorker::sftpWrite(sftp_file file, const QByteArray &buffer, const std::function<void(int bytes)> &onWritten)
-{
-    // TODO: enqueue requests.
-    // Similarly to reading we may enqueue a number of requests to mitigate the
-    // network overhead and speed up things. Serial iteration gives super poor
-    // performance.
-
-    // Break up into multiple requests in case a single request would be too large.
-    // Servers can impose an arbitrary size limit that we don't want to hit.
-    // https://bugs.kde.org/show_bug.cgi?id=404890
-    off_t offset = 0;
-    while (offset < buffer.size()) {
-        const auto length = qMin<int>(MAX_XFER_BUF_SIZE, buffer.size() - offset);
-        ssize_t bytesWritten = sftp_write(file, buffer.data() + offset, length);
-        if (bytesWritten < 0) {
-            qCDebug(KIO_SFTP_LOG) << "Failed to sftp_write" << length << "bytes."
-                                  << "- Already written (for this call):" << offset << "- Return of sftp_write:" << bytesWritten
-                                  << "- SFTP error:" << sftp_get_error(mSftp) << "- SSH error:" << ssh_get_error_code(mSession)
-                                  << "- SSH errorString:" << ssh_get_error(mSession);
-            return false;
-        } else if (onWritten) {
-            onWritten(bytesWritten);
-        }
-        offset += bytesWritten;
-    }
-    return true;
-}
-
 Result SFTPWorker::copy(const QUrl &src, const QUrl &dest, int permissions, KIO::JobFlags flags)
 {
     qCDebug(KIO_SFTP_LOG) << src << " -> " << dest << " , permissions = " << QString::number(permissions) << ", overwrite = " << (flags & KIO::Overwrite)
@@ -1612,16 +1581,19 @@ Result SFTPWorker::copy(const QUrl &src, const QUrl &dest, int permissions, KIO:
 
     if (!isSourceLocal && isDestinationLocal) { // sftp -> file
         return sftpCopyGet(src, dest.toLocalFile(), permissions, flags);
-    } else if (isSourceLocal && !isDestinationLocal) { // file -> sftp
+    }
+    if (isSourceLocal && !isDestinationLocal) { // file -> sftp
         return sftpCopyPut(dest, src.toLocalFile(), permissions, flags);
     }
 
     return Result::fail(ERR_UNSUPPORTED_ACTION);
 }
 
-Result SFTPWorker::sftpCopyGet(const QUrl &url, const QString &sCopyFile, int permissions, KIO::JobFlags flags)
+Result SFTPWorker::sftpCopyGet(const QUrl &url, const QString &sCopyFile, int permissionsMode, KIO::JobFlags flags)
 {
-    qCDebug(KIO_SFTP_LOG) << url << "->" << sCopyFile << ", permissions=" << permissions;
+    qCDebug(KIO_SFTP_LOG) << url << "->" << sCopyFile << ", permissions=" << permissionsMode;
+
+    auto permissions = posixToOptionalPerms(permissionsMode);
 
     // check if destination is ok ...
     QFileInfo copyFile(sCopyFile);
@@ -1652,20 +1624,16 @@ Result SFTPWorker::sftpCopyGet(const QUrl &url, const QString &sCopyFile, int pe
         }
     }
 
-    if (bPartExists && !bResume) // get rid of an unwanted ".part" file
+    if (bPartExists && !bResume) { // get rid of an unwanted ".part" file
         QFile::remove(sPart);
+    }
 
     // WABA: Make sure that we keep writing permissions ourselves,
     // otherwise we can be in for a surprise on NFS.
-    mode_t initialMode;
-    if (permissions != -1)
-#ifdef Q_OS_WIN
-        initialMode = permissions | static_cast<mode_t>(perms::owner_write);
-#else
-        initialMode = permissions | S_IWUSR;
-#endif
-    else
-        initialMode = 0666;
+    perms initialMode = perms::owner_read | perms::owner_write | perms::group_read | perms::group_write | perms::others_read | perms::others_write;
+    if (permissions.has_value()) {
+        initialMode = permissions.value() | perms::owner_write;
+    }
 
     // open the output file ...
     int fd = -1;
@@ -1680,7 +1648,7 @@ Result SFTPWorker::sftpCopyGet(const QUrl &url, const QString &sCopyFile, int pe
         }
         qCDebug(KIO_SFTP_LOG) << "resuming at" << offset;
     } else {
-        fd = QT_OPEN(QFile::encodeName(dest).constData(), O_CREAT | O_TRUNC | O_WRONLY, initialMode);
+        fd = QT_OPEN(QFile::encodeName(dest).constData(), O_CREAT | O_TRUNC | O_WRONLY, permsToPosix(initialMode));
     }
 
     if (fd == -1) {
@@ -2083,9 +2051,8 @@ bool SFTPWorker::GetRequest::enqueueChunks()
         if (request.id < 0) {
             if (m_pendingRequests.isEmpty()) {
                 return false;
-            } else {
-                break;
             }
+            break;
         }
 
         m_pendingRequests.enqueue(request);
@@ -2133,14 +2100,15 @@ int SFTPWorker::GetRequest::readChunks(QByteArray &data)
             }
 
             break;
-        } else if (bytesread == SSH_ERROR) {
+        }
+        if (bytesread == SSH_ERROR) {
             return -1;
         }
 
         totalRead += bytesread;
 
         if (bytesread < request.expectedLength) {
-            int rc;
+            int rc = -1;
 
             // If less data is read than expected - requeue the request
             data.resize(data.size() - (request.expectedLength - bytesread));
@@ -2185,12 +2153,12 @@ int SFTPWorker::GetRequest::readChunks(QByteArray &data)
 SFTPWorker::GetRequest::~GetRequest()
 {
     SFTPWorker::GetRequest::Request request{};
-    char buf[MAX_XFER_BUF_SIZE];
+    std::array<char, MAX_XFER_BUF_SIZE> buf{};
 
     // Remove pending reads to avoid memory leaks
     while (!m_pendingRequests.isEmpty()) {
         request = m_pendingRequests.dequeue();
-        sftp_async_read(m_file, buf, request.expectedLength, request.id);
+        sftp_async_read(m_file, buf.data(), request.expectedLength, request.id);
     }
 }
 
@@ -2259,4 +2227,67 @@ Result SFTPWorker::fileSystemFreeSpace(const QUrl &url)
     return Result::pass();
 }
 
+QCoro::Generator<SFTPWorker::ReadResponse> SFTPWorker::asyncRead(sftp_file file, size_t size)
+{
+    SFTPWorker::GetRequest request(file, size);
+
+    while (true) {
+        // Enqueue get requests
+        if (!request.enqueueChunks()) {
+            co_yield {.error = KIO::ERR_CANNOT_READ};
+            break;
+        }
+
+        QByteArray filedata;
+        const auto bytesread = request.readChunks(filedata);
+        // Read pending get requests
+        if (bytesread == -1) {
+            co_yield {.error = KIO::ERR_CANNOT_READ};
+            break;
+        }
+        if (bytesread == 0) {
+            if (file->eof) {
+                break;
+            }
+            continue;
+        }
+
+        co_yield {.filedata = filedata};
+    }
+}
+
+QCoro::Generator<SFTPWorker::WriteResponse> SFTPWorker::asyncWrite(sftp_file file, QCoro::Generator<ReadResponse> reader)
+{
+    // TODO: enqueue requests.
+    // Similarly to reading we may enqueue a number of requests to mitigate the
+    // network overhead and speed up things. Serial iteration gives super poor
+    // performance.
+
+    for (const auto &response : reader) {
+        if (response.error) {
+            co_yield {.error = response.error};
+            break;
+        }
+
+        // Break up into multiple requests in case a single request would be too large.
+        // Servers can impose an arbitrary size limit that we don't want to hit.
+        // https://bugs.kde.org/show_bug.cgi?id=404890
+        off_t offset = 0;
+        while (offset < response.filedata.size()) {
+            const auto length = qMin<int>(MAX_XFER_BUF_SIZE, response.filedata.size() - offset);
+            auto bytesWritten = sftp_write(file, response.filedata.data() + offset, length);
+            if (bytesWritten < 0) {
+                qCDebug(KIO_SFTP_LOG) << "Failed to sftp_write" << length << "bytes."
+                                      << "- Already written (for this call):" << offset << "- Return of sftp_write:" << bytesWritten
+                                      << "- SFTP error:" << sftp_get_error(mSftp) << "- SSH error:" << ssh_get_error_code(mSession)
+                                      << "- SSH errorString:" << ssh_get_error(mSession);
+                co_yield {.error = KIO::ERR_CANNOT_WRITE};
+                break;
+            }
+            co_yield {.bytes = std::make_unsigned_t<size_t>(bytesWritten)};
+            offset += bytesWritten;
+        }
+    }
+}
+
 #include "kio_sftp.moc"
